// Boid クラスを定義
// Based on http://www.openprocessing.org/visuals/?visualID=6910

var Boid = function() {

  var vector = new THREE.Vector3(),
      _acceleration,                              // 加速度
      _width = 500, _height = 500, _depth = 200,  // 動き回る範囲 (壁の位置)
      _goal,
      _neighborhoodRadius = 100,
      _maxSpeed = 4,
      _maxSteerForce = 0.1,   // (steer=舵, force=力)
      _avoidWalls = false;    // 範囲の境界 (壁) を避けるか

  this.position = new THREE.Vector3();   // 位置
  this.velocity = new THREE.Vector3();   // 速度
  _acceleration = new THREE.Vector3();   // 加速度


  //--------------------------------------------------------------------------------
  // 関数 (メソッド) の定義

  this.setGoal = function ( target ) {
    _goal = target;
  }

  this.setAvoidWalls = function ( value ) {
    _avoidWalls = value;
  }

  this.setWorldSize = function ( width, height, depth ) {
    _width = width;
    _height = height;
    _depth = depth;
  }

  // Boid の動きを 1 ステップ分進める. この関数が Main から呼び出される.
  this.run = function ( boids ) {
    if ( _avoidWalls ) {
      vector.set( - _width, this.position.y, this.position.z );
      vector = this.avoid( vector );
      vector.multiplyScalar( 5 );
      _acceleration.add( vector );

      vector.set( _width, this.position.y, this.position.z );
      vector = this.avoid( vector );
      vector.multiplyScalar( 5 );
      _acceleration.add( vector );

      vector.set( this.position.x, - _height, this.position.z );
      vector = this.avoid( vector );
      vector.multiplyScalar( 5 );
      _acceleration.add( vector );

      vector.set( this.position.x, _height, this.position.z );
      vector = this.avoid( vector );
      vector.multiplyScalar( 5 );
      _acceleration.add( vector );

      vector.set( this.position.x, this.position.y, - _depth );
      vector = this.avoid( vector );
      vector.multiplyScalar( 5 );
      _acceleration.add( vector );

      vector.set( this.position.x, this.position.y, _depth );
      vector = this.avoid( vector );
      vector.multiplyScalar( 5 );
      _acceleration.add( vector );
    }
    //else {
    //
    //  this.checkBounds();
    //
    //}

    if ( Math.random() > 0.5 ) {
      this.flock( boids );
    }
    this.move();
  }

  // (flock = 群れる)
  this.flock = function ( boids ) {
    if ( _goal ) {
      _acceleration.add( this.reach( _goal, 0.005 ) );
    }

    // ** Boid 規則の適用 **
    _acceleration.add( this.alignment( boids ) );
    _acceleration.add( this.cohesion( boids ) );
    _acceleration.add( this.separation( boids ) );
  }

  this.move = function () {
    this.velocity.add( _acceleration );

    var l = this.velocity.length();
    if ( l > _maxSpeed ) {
      this.velocity.divideScalar( l / _maxSpeed );
    }

    this.position.add( this.velocity );
    _acceleration.set( 0, 0, 0 );
  }

  this.checkBounds = function () {
    if ( this.position.x >   _width  ) this.position.x = - _width;
    if ( this.position.x < - _width  ) this.position.x =   _width;
    if ( this.position.y >   _height ) this.position.y = - _height;
    if ( this.position.y < - _height ) this.position.y =   _height;
    if ( this.position.z >   _depth  ) this.position.z = - _depth;
    if ( this.position.z < - _depth  ) this.position.z =   _depth;
  }


  //----------------------------------------------------------------------

  // 回避 (Avoid) -- target を避けるように舵をとる
  this.avoid = function ( target ) {
    var steer = new THREE.Vector3();  // (steer = 舵)
    steer.copy( this.position );
    steer.sub( target );
    steer.multiplyScalar( 1 / this.position.distanceToSquared( target ) );
    return steer;
  }

  // 反発 (Pepulse) -- target から離れる方向に加速する
  this.repulse = function ( target ) {
    var distance = this.position.distanceTo( target );

    if ( distance < 150 ) {
      var steer = new THREE.Vector3();
      steer.subVectors( this.position, target );
      steer.multiplyScalar( 0.5 / distance );
      _acceleration.add( steer );
    }
  }

  // (Reach) -- target に向かって舵をとる (向かう度合いの大きさは amount で指定)
  this.reach = function ( target, amount ) {
    var steer = new THREE.Vector3();
    steer.subVectors( target, this.position );
    steer.multiplyScalar( amount );
    return steer;
  }

  //----------------------------------------------------------------------
  // (Boid 規則)
  // 以下の 3 つの関数はすべて
  // * すべての Boid たちの配列 (boids) を引数として受け取って,
  // * それを元に計算した 力 (実際は加速度の変化量) を返す.

  // 整列 (Alignment)
  this.alignment = function ( boids ) {
    var boid,
        velSum = new THREE.Vector3(),
        count = 0;

    for ( var i = 0, il = boids.length; i < il; i++ ) {
      if ( Math.random() > 0.6 ) continue;
      boid = boids[ i ];
      distance = boid.position.distanceTo( this.position );
      if ( distance > 0 && distance <= _neighborhoodRadius ) {
        velSum.add( boid.velocity );
        count++;
      }
    }

    if ( count > 0 ) {
      velSum.divideScalar( count );
      var l = velSum.length();
      if ( l > _maxSteerForce ) {
        velSum.divideScalar( l / _maxSteerForce );
      }
    }

    return velSum;
  }

  // 結合 (Cohesion)
  this.cohesion = function ( boids ) {
    var boid, distance,
        posSum = new THREE.Vector3(),
        steer = new THREE.Vector3(),
        count = 0;

    for ( var i = 0, il = boids.length; i < il; i ++ ) {
      if ( Math.random() > 0.6 ) continue;
      boid = boids[ i ];
      distance = boid.position.distanceTo( this.position );
      if ( distance > 0 && distance <= _neighborhoodRadius ) {
        posSum.add( boid.position );
        count++;
      }
    }

    if ( count > 0 ) {
      posSum.divideScalar( count );
    }

    steer.subVectors( posSum, this.position );
    var l = steer.length();
    if ( l > _maxSteerForce ) {
      steer.divideScalar( l / _maxSteerForce );
    }

    return steer;
  }

  // 分離 (Separation)
  this.separation = function ( boids ) {
    var boid, distance,
        posSum = new THREE.Vector3(),
        repulse = new THREE.Vector3();

    for ( var i = 0, il = boids.length; i < il; i ++ ) {
      if ( Math.random() > 0.6 ) continue;
      boid = boids[ i ];
      distance = boid.position.distanceTo( this.position );
      if ( distance > 0 && distance <= _neighborhoodRadius ) {
        repulse.subVectors( this.position, boid.position );
        repulse.normalize();
        repulse.divideScalar( distance );
        posSum.add( repulse );
      }
    }
    return posSum;
  }

}
